window.vteapp-window scrolledwindow undershoot.top,
window.vteapp-window scrolledwindow undershoot.bottom,
window.vteapp-window scrolledwindow overshoot.top,
window.vteapp-window scrolledwindow overshoot.bottom
{
  background: none;
}

